/*****************************************************************************/
/* Project:     Fractals                                                        
/* Author:      Eugene Sirota                                                   
/* Created:     2008                                                            
/*****************************************************************************/

using Fractals.Core;
using System.Drawing; 

namespace Fractals.Geometry
{
	public delegate void FractalFragmentCalculatedHandler(LinearFractal sender, FractalFragmentCalculatedEventArgs ea);		

	public class LinearFractal : PositionedObject
	{
		private readonly Shape baseFragment;
		private readonly Line[] baseLines;

		public event FractalFragmentCalculatedHandler FractalFragmentCalculatedEvent;

        public LinearFractal(Line baseLine, Shape baseFragment)
            : this(new[] { baseLine }, baseFragment)
        {
        }

		public LinearFractal(Line[] baseLines, Shape baseFragment)
		{
            this.baseFragment = baseFragment.Clone() as Shape;
			this.baseLines = baseLines.DeepClone(); 
		}

        public override object Clone()
        {
            return new LinearFractal(baseLines, baseFragment);
        }

		public override void Rotate(float angle)
		{
			foreach (var line in baseLines)
				line.Rotate(angle);
			baseFragment.Rotate(angle);
		}

		public override void Scale(float scale)
		{
			baseFragment.Scale(scale); 
			foreach (var baseLine in baseLines)
				baseLine.Scale(scale);					
		}

		public void Calculate(int deep)
		{
            if (deep < 1)
                return;

			foreach (var baseLine in baseLines)
				CalculateInternal(baseLine, baseFragment, 1, deep);
		}

		public void CalculateInternal(Line baseLine, Shape fragment,int currentDeep, int deep)
		{			
		    if (FractalFragmentCalculatedEvent != null)
                FractalFragmentCalculatedEvent(this, new FractalFragmentCalculatedEventArgs((Line)baseLine.Clone(), (Shape)fragment.Clone(), currentDeep));

            if (currentDeep == deep)
				return;

			foreach (var line in fragment.Lines)
			{
				var nextFragment = GetNextFragment(baseLine, fragment, line);
				if (nextFragment != null)
                    CalculateInternal(line, nextFragment, currentDeep + 1, deep); 
			}
		}

		private static Shape GetNextFragment(Line baseLine, Shape fragment, Line nextBaseLine)
		{
            if (baseLine.Length == 0 || nextBaseLine.Length == 0)
                return null;
            
            var vectorA = baseLine.Vector;
		    var vectorB = nextBaseLine.Vector;

            var scalarProd = vectorA.X * vectorB.X + vectorA.Y * vectorB.Y;
            var determinant = vectorA.X * vectorB.Y - vectorA.Y * vectorB.X;

		    var moduleA = baseLine.Length;
		    var moduleB = nextBaseLine.Length;
            
            var cosa = scalarProd / (moduleA * moduleB);
            var sina = determinant / (moduleA * moduleB);

			var scale = moduleB / moduleA;

            var centerA = baseLine.Center;
		    var centerB = nextBaseLine.Center;
			
			var nextFragmentLines = new Line[fragment.Lines.Length];
            for (var i = 0; i < fragment.Lines.Length; i++)
            {
                var line = fragment.Lines[i];
				var nextLine = (Line)line.Clone();
				nextLine.Move(-centerA.X, -centerA.Y);

                var sx = cosa * nextLine.Start.X - sina * nextLine.Start.Y;
                var sy = sina * nextLine.Start.X + cosa * nextLine.Start.Y;
                nextLine.Start = new PointF(sx, sy);

                var ex = cosa * nextLine.End.X - sina * nextLine.End.Y;
                var ey = sina * nextLine.End.X + cosa * nextLine.End.Y;
                nextLine.End = new PointF(ex, ey);
				
                nextLine.Scale(scale);
				nextLine.Move(centerB);
				
				nextFragmentLines[i] = nextLine;
            }
		    return new Shape(nextFragmentLines);		   
		}		
	}
}
